//	Medianblur version 0.3
//  copyright (C) Tonny Petersen (tsp@person.dk) 2004

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .




class MedianBlur : public GenericVideoFilter {   
public:
	
	MedianBlur(PClip,int,int,int,bool,IScriptEnvironment* env);
protected:
	
	virtual void CalcFrame(PVideoFrame&,PVideoFrame&,int);
	unsigned int _radius;
	const int _radiusy;
	const int _radiusu;
	const int _radiusv;
	const bool calcborder;
	unsigned int count[256];
	unsigned int median;
	unsigned int w;
	unsigned int h;
	unsigned int N;
	unsigned int x;
	unsigned int n;
	
	virtual BYTE radixmedianInit(const BYTE*,unsigned int);
	virtual BYTE radixmedianAdd(const BYTE*);
	virtual void radixmedianRemove(const BYTE*);
	virtual BYTE radixmedianInitXY(const BYTE*,unsigned int,unsigned int,unsigned int);
	virtual BYTE radixmedianAddXY(const BYTE*,unsigned int);
	virtual BYTE radixmedianRemoveXY(const BYTE*,unsigned int);
	virtual void radixmedianRemoveXY(const BYTE*);
	
	
};

class MedianBlurMMX : public MedianBlur {
	public:
		PVideoFrame __stdcall GetFrame(int, IScriptEnvironment*);
		MedianBlurMMX(PClip,int,int,int,bool,IScriptEnvironment* env);
	private:
		void CalcFrameMMX(PVideoFrame&,PVideoFrame&,int);

		void CalcMedian(unsigned int x,unsigned int y,unsigned int pitch,const BYTE *srcp,BYTE *dstp);
		void CalcMedian16(unsigned int x,unsigned int y,unsigned int pitch,const BYTE *srcp,BYTE *dstp);
		void CalcBatcherMedian16(unsigned int x,unsigned int y,unsigned int pitch,const BYTE *srcp,BYTE *dstp);
		
	
	
};

class MedianBlurWeight : public MedianBlur {
public:
	
	PVideoFrame __stdcall GetFrame(int, IScriptEnvironment*);
	MedianBlurWeight(PClip cl1,int rady,int radu,int radv,int wy,int wu,int wv,bool cb,IScriptEnvironment* env);
protected:
	
	unsigned int _weight;
	unsigned int _weighty;
	unsigned int _weightu;
	unsigned int _weightv;
	unsigned int spitch;
	unsigned int _offset;
	const BYTE* sp;

	virtual BYTE radixmedianInit (const BYTE *source,unsigned int pitch);
	virtual void radixmedianRemove(const BYTE*);
	virtual BYTE radixmedianAdd(const BYTE*);
	virtual	BYTE radixmedianInitXY(const BYTE*,unsigned int,unsigned int,unsigned int);
	virtual BYTE radixmedianAddXY(const BYTE*,unsigned int);
	virtual BYTE radixmedianRemoveXY(const BYTE*,unsigned int);
};

class MedianBlurBin : public MedianBlur {

public:
	MedianBlurBin(PClip,int,int,int,int,int,int,bool,IScriptEnvironment* env);
	PVideoFrame __stdcall GetFrame(int, IScriptEnvironment*);
protected:	
	unsigned int _thres;
	unsigned int _thresy;
	unsigned int _thresu;
	unsigned int _thresv;
	__int64 _t;
	bool iSSE;


	virtual BYTE radixmedianInit(const BYTE*,unsigned int);
	virtual BYTE radixmedianAdd(const BYTE*);
	virtual void radixmedianRemove(const BYTE*);
	virtual BYTE radixmedianInitXY(const BYTE*,unsigned int,unsigned int,unsigned int);
	virtual BYTE radixmedianAddXY(const BYTE*,unsigned int);
	virtual BYTE radixmedianRemoveXY(const BYTE*,unsigned int);
	virtual void radixmedianRemoveXY(const BYTE*);
	void tinit();
	int arasm(int,const BYTE*);
	int iasm(int x,int y,const BYTE *source,unsigned int pitch);

};

class MedianBlurTemp : public MedianBlur {

public:
	MedianBlurTemp(PClip,int,int,int,int,bool,bool,bool,int,int,int,int,int,int,bool,bool,IScriptEnvironment* env);
	~MedianBlurTemp();
	PVideoFrame __stdcall GetFrame(int, IScriptEnvironment*);
protected:
	void CalcFrame(PVideoFrame&,PVideoFrame&,int,int,IScriptEnvironment*);
	bool MC;
	//int blksize,pel,lambda, thSCD1, thSCD2,searchparam;
	bool scenechange;
	bool markscenechange;
	//bool spfull;
	unsigned int _trad;
	unsigned int twidth;
	const BYTE **srcpa;
	PVideoFrame *srca;
	PClip *clipa;
	void Alignpitch(PVideoFrame &src1, IScriptEnvironment* env);

		
	virtual BYTE radixmedianInit(unsigned int offset,unsigned int pitch);
	virtual void radixmedianRemove(unsigned int offset);
	virtual BYTE radixmedianAdd(unsigned int offset);
	virtual BYTE radixmedianInitXY(unsigned int offset,unsigned int,unsigned int,unsigned int);
	virtual BYTE radixmedianAddXY(unsigned int offset,unsigned int);
	virtual BYTE radixmedianRemoveXY(unsigned int offset,unsigned int);
	virtual void radixmedianRemoveXY(unsigned int offset);

};



class ML3Dex : public GenericVideoFilter {

public:
	PVideoFrame __stdcall GetFrame(int, IScriptEnvironment*);
	ML3Dex(PClip,bool,int,int,int,bool,int,int,int,int,int,int,bool,bool,IScriptEnvironment* env);
protected:
	bool iSSE;
	int _y,_u,_v;
	bool scenechange;
	bool markscenechange;
	//bool spfull;
	bool MC;
	//int blksize,pel,lambda, thSCD1, thSCD2,searchparam;
	const BYTE *srcpa[3];
	PVideoFrame srca[3];
	PClip clipa[3];
	BYTE medianmedianbuffer[5][16];
	BYTE medianbuffer[11][16];
	void CalcPixel(unsigned int _offset,BYTE* dstp,unsigned int pitch);
	void CalcFrame(int plane,PVideoFrame &src,PVideoFrame &dst);
	void Alignpitch(PVideoFrame &src1, IScriptEnvironment* env);
};
	





/*
* This Quickselect routine is based on the algorithm described in
* "Numerical recipes in C", Second Edition,
* Cambridge University Press, 1992, Section 8.5, ISBN 0-521-43108-5
* This code by Nicolas Devillard - 1998. Public domain.
*/
#define ELEM_SWAP(a,b) { register unsigned int t=(a);(a)=(b);(b)=t; }
unsigned int quick_select(unsigned int arr[], int n)
{
int low, high ;
int median;
int middle, ll, hh;
low = 0 ; high = n-1 ; median = (low + high) / 2;
for (;;) {
if (high <= low) /* One element only */
return arr[median] ;
if (high == low + 1) { /* Two elements only */
if (arr[low] > arr[high])
ELEM_SWAP(arr[low], arr[high]) ;
return arr[median] ;
}
/* Find median of low, middle and high items; swap into position low */
middle = (low + high) / 2;
if (arr[middle] > arr[high]) ELEM_SWAP(arr[middle], arr[high]) ;
if (arr[low] > arr[high]) ELEM_SWAP(arr[low], arr[high]) ;
if (arr[middle] > arr[low]) ELEM_SWAP(arr[middle], arr[low]) ;
/* Swap low item (now in position middle) into position (low+1) */
ELEM_SWAP(arr[middle], arr[low+1]) ;
/* Nibble from each end towards middle, swapping items when stuck */
ll = low + 1;
hh = high;
for (;;) {
do ll++; while (arr[low] > arr[ll]) ;
do hh--; while (arr[hh] > arr[low]) ;
if (hh < ll)
break;
ELEM_SWAP(arr[ll], arr[hh]) ;
}
/* Swap middle item (in position low) back into correct position */
ELEM_SWAP(arr[low], arr[hh]) ;
/* Re-set active partition */
if (hh <= median)
low = ll;
if (hh >= median)
high = hh - 1;
}
}
#undef ELEM_SWAP
